//  Vamos Automotive Simulator
//  Copyright (C) 2001--2004 Sam Varner
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <GL/glut.h>

#include <vamos/body/Gl_Car.h>
#include <vamos/geometry/Texture_Image.h>
#include <vamos/track/Strip_Track.h>
#include <vamos/world/Gl_World.h>
#include <vamos/world/Sounds.h>

using Vamos_Geometry::Three_Vector;

bool get_options (int argc, char* argv [],
                  std::string& car_file,
                  std::string& track_file,
                  std::string& world_file,
                  std::string& controls_file,
                  int& number_of_opponents,
                  int& focused_car,
                  bool& map_mode)
{
  const std::string car_option = "--car=";
  const std::string track_option = "--track=";
  const std::string world_option = "--world=";
  const std::string controls_option = "--controls=";
  const std::string opponents_option = "--opponents=";
  const std::string focused_car_option = "--focused-car=";

  car_file = "default-car";
  track_file = "default-track";
  world_file = "default-world";
  controls_file = "default-controls";
  number_of_opponents = 0;
  focused_car = 0;

  map_mode = false;
  bool error = false;

  for (int i = 1; i < argc; i++)
    {
      std::string arg = argv [i];

      if ((arg == "-m") || (arg == "--map"))
        {
          map_mode = true;
        }
      else if (arg == "-t")
        {
          if (argc >= i + 1)
              track_file = argv [++i];
          else
              error = true;
        }
      else if (arg.substr (0, track_option.size ()) == track_option)
        {
          track_file = argv [i] + track_option.size ();
        }
      else if (arg == "-c")
        {
          if (argc >= i + 1)
              car_file = argv [++i];
          else
              error = true;
        }
      else if (arg.substr (0, car_option.size ()) == car_option)
        {
          car_file = argv [i] + car_option.size ();
        }
      else if (arg == "-w")
        {
          if (argc >= i + 1)
              world_file = argv [++i];
          else
              error = true;
        }
      else if (arg.substr (0, world_option.size ()) == world_option)
        {
          world_file = argv [i] + world_option.size ();
        }
      else if (arg == "-a")
        {
          if (argc >= i + 1)
              controls_file = argv [++i];
          else
              error = true;
        }
      else if (arg.substr (0, controls_option.size ()) == controls_option)
        {
          controls_file = argv [i] + controls_option.size ();
        }
      else if (arg == "-o")
        {
          if (argc >= i + 1)
            number_of_opponents = atoi (argv [++i]);
          else
            error = true;
        }
      else if (arg.substr (0, opponents_option.size ()) == opponents_option)
        {
          number_of_opponents = atoi (argv [i] + opponents_option.size ());
        }
      else if (arg == "-f")
        {
          if (argc >= i + 1)
            focused_car = atoi (argv [++i]);
          else
            error = true;
        }
      else if (arg.substr (0, focused_car_option.size ()) 
               == focused_car_option )
        {
          focused_car = atoi (argv [i] + focused_car_option.size ());
        }
      else
        error = true;
    }

  return error;
}

int main (int argc, char* argv [])
{
  std::string car_file;
  std::string track_file;
  std::string world_file;
  std::string controls_file;
  int number_of_opponents;
  int focused_car;

  bool map_mode;
  bool error = get_options (argc, argv,
                            car_file, track_file, 
                            world_file, controls_file,
                            number_of_opponents,
                            focused_car,
                            map_mode);

  if (error)
    {
      std::cerr << "Usage: vamos "
                << "[-m|--map] "
                << "[[-t|--track=] TRACK_FILE] "
                << "[[-c|--car=] CAR_FILE] "
                << "[[-w|--world=] WORLD_FILE] "
                << "[[-a|--controls=] CONTROLS_FILE] "
                << "[[-o|--opponents=] NUMBER_OF_OPPONENTS] "
                << "[[-f|--focused-car=] FOCUSED_CAR_INDEX] "
                << std::endl;
      std::exit (EXIT_FAILURE);
    }

  track_file = "tracks/" + track_file + ".xml";
  car_file = "cars/" + car_file + ".xml";
  world_file = "worlds/" + world_file + ".xml";
  controls_file = "controls/" + controls_file + ".xml";


  Vamos_World::Atmosphere air (1.2, Three_Vector (0.0, 0.0, 0.0));

  // Check the source directory for data files.
 std::string data_dir = "../data/";
  try
    {
      Vamos_Geometry::Texture_Image test_tex (data_dir + "textures/wall.png");
    }
  catch (Vamos_Geometry::Missing_Texture_File)
    {
      // They're not there.  Maybe this is an installed version.

      // How I can I get this automatically set to the configured data
      // directory?
      data_dir = "/usr/local/share/vamos/";
      try
        {
          Vamos_Geometry::Texture_Image 
            test_tex (data_dir + "textures/wall.png");
        }
      catch (Vamos_Geometry::Missing_Texture_File)
        {
          std::cerr << "Couldn't find the data direcory  ../data, or "
                    << "/usr/local/share/vamos/."
                    << std::endl;
          std::exit (EXIT_FAILURE);
        }
    }

  Vamos_Track::Strip_Track road;

  Vamos_World::Sounds* sounds = 0;
  if (!map_mode)
    {
      sounds = new Vamos_World::Sounds ();
      try
        {
          sounds->read (data_dir + "sounds/", "default-sounds.xml");
        }
      catch (Vamos_Geometry::XML_Exception& error)
        {
          std::cerr << error.message () << std::endl;
          std::exit (EXIT_FAILURE);
        }
    }

  Vamos_World::Gl_World world (argc, argv, &road, &air, sounds);

  Vamos_Body::Gl_Car* car = 0;
  if (!map_mode)
    {
      try
        {
          car = new Vamos_Body::Gl_Car (Three_Vector (11.0, 0.0, 0.6));
          car->read (data_dir, car_file);
          car->chassis ().translate (Three_Vector (0.0, 0.0, 
                                                   -car->chassis ().lowest_contact_position () + 0.5));
          car->start_engine ();
          world.add_car (car);

          for (int i = 0; i < number_of_opponents; i++)
            {
              car = new Vamos_Body::Gl_Car (Three_Vector (11.0, 0.0, 0.6));
              car->read (data_dir, car_file);
              car->chassis ().translate (Three_Vector (10.0 * (i + 1), 0.0, 
                                                       -car->chassis ().lowest_contact_position () + 0.5));
              car->start_engine ();
              world.add_car (car);
            }

          world.set_focused_car (focused_car);
        }
      catch (Vamos_Geometry::XML_Exception& error)
        {
          std::cerr << error.message () << std::endl;
          std::exit (EXIT_FAILURE);
        }
    }

  try
    {
      road.read (data_dir, track_file);
    }
  catch (Vamos_Geometry::XML_Exception& error)
    {
      std::cerr << error.message () << std::endl;
      std::exit (EXIT_FAILURE);
    }

  try
    {
      world.read (data_dir, world_file, controls_file);
    }
  catch (Vamos_Geometry::XML_Exception& error)
    {
      std::cerr << error.message () << std::endl;
      std::exit (EXIT_FAILURE);
    }
  world.start ();

  delete sounds;

  return EXIT_SUCCESS;
}
