//  Controls.cc - a class for handling control events.
//
//	Vamos Automotive Simulator
//  Copyright (C) 2003 Sam Varner
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <vamos/world/Controls.h>
#include <vamos/geometry/Conversions.h>

#include <GL/glut.h>
#include <plib/js.h>

#include <iostream>
#undef max

Vamos_World::Controls::
Callback::Callback (Gl_World* obj, Callback_Func func, 
					double d1, double d2, double d3, double d4)
  : object (obj), 
	function (func), 
	data1 (d1), 
	data2 (d2), 
	data3 (d3),
	data4 (d4) 
{
}

//-----------------------------------------------------------------------------
void Vamos_World::Controls::
Keyboard::bind_key_down (int key, 
						 Gl_World* obj, 
						 Callback_Func func,
						 double time)
{
  m_key_down_map.push_back (std::make_pair (key, Callback (obj, func, time)));
} 

void Vamos_World::Controls::
Keyboard::bind_key_up (int key, 
					   Gl_World* obj, 
					   Callback_Func func,
					   double time)
{
  m_key_up_map.push_back (std::make_pair (key, Callback (obj, func, time)));
} 

void Vamos_World::Controls::
Keyboard::call_key_callbacks (const Callback_List& callbacks, 
							  int match, double value) const
{
  bool done_calling = false;
  for (Callback_List::const_iterator it = callbacks.begin ();
	   (it != callbacks.end ()) && !done_calling;
	   it++)
	{
	  if (it->first == match)
		{
		  const Callback& c = it->second;
		  done_calling = ((c.object)->*(c.function)) (value, c.data1);
		}
	}
}

void Vamos_World::Controls::
Keyboard::key_press (int key)
{
  call_key_callbacks (m_key_down_map, key, 1.0);
}

void Vamos_World::Controls::
Keyboard::key_release (int key)
{
  call_key_callbacks (m_key_up_map, key, 0.0);
}

//-----------------------------------------------------------------------------
void Vamos_World::Controls::
Continuous_Control::bind_button_down (int button, 
									  Gl_World* obj, 
									  Callback_Func func,
									  double time)
{
  m_button_down_map.push_back (std::make_pair (translate_button (button), 
											   Callback (obj, func, time)));
}

void Vamos_World::Controls::
Continuous_Control::bind_button_up (int button, 
									Gl_World* obj, 
									Callback_Func func,
									double time)
{
  m_button_up_map.push_back (std::make_pair (translate_button (button), 
											 Callback (obj, func, time)));
}

void Vamos_World::Controls::
Continuous_Control::call_button_callbacks (const Callback_List& callbacks,
										   int match, double value) const
{
  call_button_callbacks (callbacks, match, 0, value);
}

void Vamos_World::Controls::
Continuous_Control::call_button_callbacks (const Callback_List& callbacks,
										   int match, 
										   int dont_match, 
										   double value) const
{
  bool done_calling = false;
  for (Callback_List::const_iterator it = callbacks.begin ();
	   (it != callbacks.end ()) && !done_calling;
	   it++)
	{
	  if (button_matches (it->first, match, dont_match))
		{
		  const Callback& c = it->second;
		  done_calling = ((c.object)->*(c.function)) (value, c.data1);
		}
	}
}

//-----------------------------------------------------------------------------
Vamos_World::Controls::
Joystick::Joystick ()
  : mp_js_joystick (new jsJoystick),
	m_joystick_is_present (false),
	m_num_axes (0),
	mp_js_axes (0),
	m_last_buttons (0)
{
  // If the joystick is not not working, then it's present.
  if (mp_js_joystick->notWorking () == JS_FALSE)
	{
	  m_joystick_is_present = true;

	  m_num_axes = mp_js_joystick->getNumAxes ();
	  mp_js_axes = new float [m_num_axes];
	}  
}

Vamos_World::Controls::
Joystick::~Joystick ()
{
  if (m_joystick_is_present)
	{
	  delete [] mp_js_axes;
	}
  delete mp_js_joystick;
}

bool Vamos_World::Controls::
Joystick::button_matches (int button, int match, int dont_match) const
{
  return ((button & match) && !(button & dont_match));
}

void Vamos_World::Controls::
Joystick::bind_axis (int axis, 
					 Gl_World* obj, 
					 Callback_Func func, 
					 double factor,
					 double offset,
					 double minimum,
					 double deadband)
{
  if (!m_joystick_is_present) return;

  if (axis > m_num_axes)
	{
	  throw Bad_Axis (axis, m_num_axes);
	}
  m_motion_map.push_back 
	(std::make_pair (axis, Callback (obj, func, factor, offset, minimum)));
  mp_js_joystick->setDeadBand (axis - 1, deadband);
}

void Vamos_World::Controls::
Joystick::call_axis_callbacks (const Callback_List& callbacks) const
{
  bool done_calling = false;
  for (Callback_List::const_iterator it = callbacks.begin ();
	   (it != callbacks.end ()) && !done_calling;
	   it++)
	{
	  assert (it->first <= m_num_axes);
	  const Callback& c = it->second;
	  double value = std::max (mp_js_axes [it->first - 1] * c.data1 + c.data2,
							   c.data3);
	  done_calling = ((c.object)->*(c.function)) (value, 0.0); 
	}
}

void Vamos_World::Controls::
Joystick::update ()
{
  if (!m_joystick_is_present) return;

  int buttons;
  mp_js_joystick->read (&buttons, mp_js_axes);
  call_button_callbacks (m_button_down_map, buttons, m_last_buttons, 1.0);
  call_button_callbacks (m_button_up_map, m_last_buttons, buttons, 0.0);
  m_last_buttons = buttons;
  call_axis_callbacks (m_motion_map);
}

//-----------------------------------------------------------------------------
bool Vamos_World::Controls::
Mouse::button_matches (int button, int match, int dont_match) const
{
  return (button == match);
}

void Vamos_World::Controls::
Mouse::bind_motion (int axis,
					Gl_World* obj, 
					Callback_Func func, 
					double min,
					double min_position,
					double max,
					double max_position)
{
  m_motion_map.push_back 
	(std::make_pair (axis, Callback (obj, func, 
									 min, min_position, max, max_position)));
}

double 
interp (double x, double x1, double y1, double x2, double y2)
{
  const double slope = double (y2 - y1) / (x2 - x1);
  return slope * (x - x1) + y1;
}

double Vamos_World::Controls::
Mouse::get_position (int dir, int x, int y) const
{
  double position = 0.0;
  switch (Direction (dir))
	{
	case X:
	  position = 2.0 * x / m_window_width - 1.0;
	  break;
	case Y:
	  position = -2.0 * y / m_window_height + 1.0;
	  break;
	default:
	  assert (false);
	};
  return position;
}

void Vamos_World::Controls::
Mouse::call_motion_callbacks (const Callback_List& callbacks,
							  int x, int y) const
{
  bool done_calling = false;
  for (Callback_List::const_iterator it = callbacks.begin ();
	   (it != callbacks.end ()) && !done_calling;
	   it++)
	{
	  double position = get_position (it->first, x, y);

	  const Callback& c = it->second;
	  const double min_pos = c.data2;
	  const double max_pos = c.data4;
	  position = (min_pos < max_pos)
		? Vamos_Geometry::clip (position, min_pos, max_pos)
		: Vamos_Geometry::clip (position, max_pos, min_pos);

	  const double min_value = c.data1;
	  const double max_value = c.data3;
	  const double value = 
		interp (position, min_pos, min_value, max_pos, max_value);

	  done_calling = ((c.object)->*(c.function)) (value, 0.0); 
	}
}

void Vamos_World::Controls::
Mouse::update_button (int button, int state)
{
  switch (state)
	{
	case GLUT_DOWN:
	  call_button_callbacks (m_button_down_map, button, 1.0);
	  break;
	case GLUT_UP:
	  call_button_callbacks (m_button_up_map, button, 0.0);
	  break;
	default:
	  assert (false);
	};
}

void Vamos_World::Controls::
Mouse::update_motion (int x, int y)
{
  call_motion_callbacks (m_motion_map, x, y);
}

void Vamos_World::Controls::
Mouse::update_window_size (int width, int height)
{
  m_window_width = width;
  m_window_height = height;
}

//-----------------------------------------------------------------------------
Vamos_World::
Controls::Controls () :
  mp_keyboard (new Keyboard),
  mp_joystick (new Joystick),
  mp_mouse (new Mouse),
  m_auto_clutch (true)
{
}

Vamos_World::
Controls::~Controls ()
{
  delete mp_keyboard;
  delete mp_joystick;
  delete mp_mouse;
}

// Key Bindings
void Vamos_World::
Controls::bind_key_down (int key, 
						 Gl_World* obj, 
						 Callback_Func func,
						 double time)
{
  mp_keyboard->bind_key_down (key, obj, func, time);
} 

void Vamos_World::
Controls::bind_key_up (int key, 
					   Gl_World* obj, 
					   Callback_Func func,
					   double time)
{
  mp_keyboard->bind_key_up (key, obj, func, time);
} 

void Vamos_World::
Controls::bind_special_key_down (int key, 
								 Gl_World* obj, 
								 Callback_Func func,
								 double time)
{
  mp_keyboard->bind_key_down (translate_special_key (key), obj, func, time);
} 

void Vamos_World::
Controls::bind_special_key_up (int key, 
							   Gl_World* obj, 
							   Callback_Func func,
							   double time)
{
  mp_keyboard->bind_key_up (translate_special_key (key), obj, func, time);
} 

// Joystick Button Bindings
void Vamos_World::
Controls::bind_button_down (int button, 
							Gl_World* obj, 
							Callback_Func func,
							double time)
{
  mp_joystick->bind_button_down (button, obj, func, time);
}

void Vamos_World::
Controls::bind_button_up (int button, 
						  Gl_World* obj, 
						  Callback_Func func, 
						  double time)
{
  mp_joystick->bind_button_up (button, obj, func, time);
}

void Vamos_World::
Controls::bind_axis (int axis, 
					 Gl_World* obj, 
					 Callback_Func func, 
					 double factor,
					 double offset,
					 double minimum,
					 double deadband)
{
  mp_joystick->bind_axis (axis, obj, func, factor, offset, minimum, deadband);
}

// Mouse Bindings
void Vamos_World::
Controls::bind_mouse_button_down (int button, 
								  Gl_World* obj, 
								  Callback_Func func,
								  double time)
{
  mp_mouse->bind_button_down (button, obj, func, time);
}

void Vamos_World::
Controls::bind_mouse_button_up (int button, 
								Gl_World* obj, 
								Callback_Func func,
								double time)
{
  mp_mouse->bind_button_up (button, obj, func, time);
}

void Vamos_World::
Controls::bind_mouse_motion (int axis,
							 Gl_World* obj, 
							 Callback_Func func, 
							 double factor,
							 double offset,
							 double minimum,
							 double deadband)
{
  mp_mouse->bind_motion (axis, obj, func, factor, offset, minimum, deadband);
}

// Glut Keyboard Event Callbacks
void Vamos_World::
Controls::key_press (unsigned char key, int, int)
{
  mp_keyboard->key_press (key);
}

void Vamos_World::
Controls::key_release (unsigned char key, int, int)
{
  mp_keyboard->key_release (key);
}

void Vamos_World::
Controls::special_key_press (int key, int, int)
{
  mp_keyboard->key_press (translate_special_key (key));
}

void Vamos_World::
Controls::special_key_release (int key, int, int)
{
  mp_keyboard->key_release (translate_special_key (key));
}

// Glut Mouse Event Callbacks
void Vamos_World::
Controls::mouse_motion (int x, int y)
{
  mp_mouse->update_motion (x, y);
}

void Vamos_World::
Controls::mouse_button (int button, int state, int, int)
{
  mp_mouse->update_button (button, state);
}

void Vamos_World::
Controls::set_window_size (int width, int height)
{
  mp_mouse->update_window_size (width, height);
}

void Vamos_World::
Controls::update ()
{
  mp_joystick->update ();
}
