//  Vamos Automotive Simulator
//  Copyright (C) 2001--2004 Sam Varner
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <vamos/body/Car.h>
#include <vamos/body/Fuel_Tank.h>
#include <vamos/body/Wheel.h>
#include <vamos/geometry/Conversions.h>
#include <vamos/geometry/Gl_Texture_Image.h>
#include <vamos/track/Track.h>
#include <vamos/world/Gl_World.h>
#include <vamos/world/Sounds.h>
#include <vamos/world/Controls.h>

#include <GL/glut.h>
#include <plib/js.h>

#include <cmath>
#include <iomanip>
#include <sstream>
#include <string>
#include <cstdlib>
#include <algorithm>
#include <cctype>
#undef min

using namespace Vamos_Geometry;

// Global signals to be connected to the callback functions through proxies.
namespace Vamos_World
{
  SigC::Signal0 <void> display_signal;
  SigC::Signal2 <void, int, int> reshape_signal;
  SigC::Signal3 <void, unsigned char, int, int> key_signal;
  SigC::Signal3 <void, unsigned char, int, int> key_up_signal;
  SigC::Signal3 <void, int, int, int> special_key_signal;
  SigC::Signal3 <void, int, int, int> special_key_up_signal;
  SigC::Signal2 <void, int, int> mouse_motion_signal;
  SigC::Signal4 <void, int, int, int, int> mouse_button_signal;
  SigC::Signal0 <void> idle_signal;
  SigC::Signal0 <void> timer_signal;
}

// Define the global callback proxies.
void Vamos_World::
display_proxy ()
{
  display_signal (); 
}
void Vamos_World::
reshape_proxy (int width, int height)
{
  reshape_signal (width, height); 
}
void Vamos_World::
key_proxy (unsigned char key, int x, int y)
{
  key_signal (key, x, y); 
}
void Vamos_World::
key_up_proxy (unsigned char key, int x, int y)
{
  key_up_signal (key, x, y); 
}
void Vamos_World::
special_key_proxy (int key, int x, int y)
{
  special_key_signal (key, x, y); 
}
void Vamos_World::
special_key_up_proxy (int key, int x, int y)
{
  special_key_up_signal (key, x, y);
}
void Vamos_World::
mouse_motion_proxy (int x, int y)
{
  mouse_motion_signal (x, y);
}
void Vamos_World::
mouse_button_proxy (int button, int state, int x, int y)
{
  mouse_button_signal (button, state, x, y);
}
void Vamos_World::
animate_proxy ()
{
  idle_signal (); 
}
void Vamos_World::
timer_proxy (int value)
{
  timer_signal ();
}

namespace Vamos_World
{
  int translate_key (std::string key_name, bool* special);
  int mouse_button_to_control (std::string button);
  int mouse_direction_to_control (std::string direction);
}

//-----------------------------------------------------------------------------
namespace Vamos_World
{
  std::string format_time (double time);
  std::string format_time_difference (double time);
}

std::string Vamos_World::
format_time (double time)
{
  if (time == 0.0) return "";

  int minutes = int (time / 60.0);
  int seconds = int (time) % 60;
  int milliseconds = int (1000 * (time - floor (time)));

  std::ostringstream os;
  os << minutes << ':' << std::setfill ('0') << std::setw (2) 
     << seconds << '.' << milliseconds;
  return os.str ();
}

std::string Vamos_World::
format_time_difference (double delta_time)
{
  if (delta_time == 0.0) return "";

  std::ostringstream os;
  if (delta_time > 0.0)
    {
      os << '+';
    }
  os << std::setprecision (3) << delta_time;
  return os.str ();
}

//-----------------------------------------------------------------------------
Vamos_World::
Gl_Window::Gl_Window (int width, int height, const char* name)
{
  glutInitWindowSize (width, height);
  m_id = glutCreateWindow (name);
  resize (width, height);
}

Vamos_World::
Gl_Window::~Gl_Window ()
{
  glutDestroyWindow (m_id);
}

void Vamos_World::
Gl_Window::resize (int width, int height)
{
  m_width = width;
  m_height = height;
  m_aspect = (m_height == 0) ? 1.0 : double (m_width) / m_height; 
  glViewport (0, 0, m_width, m_height);
}

//-----------------------------------------------------------------------------
Vamos_World::
Timer::Timer (double interval_to_average_over) :
  m_timeout (int (1000.0 * interval_to_average_over)),
  m_max_time_step (0.1),
  m_frames (1)
{
  start ();
}

void Vamos_World::
Timer::start ()
{
  glutTimerFunc (m_timeout, timer_proxy, m_timeout);
}

void Vamos_World::
Timer::update ()
{
  m_time_step = (m_timeout / 1000.0) / m_frames;

  // Cap the time step at 0.03 second, 33 frames/s.  If it gets much longer
  // the simulation becomes unstable.
  m_time_step = std::min (m_time_step, 0.03);

  start ();
  m_frames = 1;
}

//-----------------------------------------------------------------------------
Vamos_World::
Gl_World::Gl_World (int argc, char** argv,
                    Vamos_Track::Strip_Track* track, 
                    Atmosphere* atmosphere,
                    Sounds* sounds)
  : World (track, atmosphere),
    mp_sounds (sounds),
    mp_controls (new Controls ()),
    mp_timer (0),
    mp_window (0),
    m_view (MAP_VIEW)
{
  initialize_graphics (&argc, argv);
  mp_window = new Gl_Window (900, 600, argv [0]);
  mp_timer = new Timer (1.0);
  set_callbacks ();
  set_attributes ();
  set_paused (true);
}

Vamos_World::Gl_World::
~Gl_World ()
{
  delete mp_window;
  delete mp_timer;
  delete mp_controls;
  delete mp_sounds;
}

void Vamos_World::
Gl_World::initialize_graphics (int* argc, char** argv)
{
  glutInit (argc, argv);
  glutInitDisplayMode (GLUT_DEPTH | GLUT_DOUBLE | GLUT_STENCIL | GLUT_RGBA);
}

void Vamos_World::
Gl_World::set_callbacks ()
{
  display_signal.connect (SigC::slot (*this, &Gl_World::display));
  reshape_signal.connect (SigC::slot (*this, &Gl_World::reshape));
  idle_signal.connect (SigC::slot (*this, &Gl_World::animate));

  timer_signal.connect (SigC::slot (*mp_timer, &Timer::update));

  key_signal.connect (SigC::slot (*mp_controls, &Controls::key_press));
  key_up_signal.connect (SigC::slot (*mp_controls, &Controls::key_release));
  special_key_signal.connect (SigC::slot (*mp_controls, 
                                          &Controls::special_key_press));
  special_key_up_signal.connect (SigC::slot (*mp_controls, 
                                             &Controls::special_key_release));
  mouse_motion_signal.connect (SigC::slot (*mp_controls,
                                           &Controls::mouse_motion));
  mouse_button_signal.connect (SigC::slot (*mp_controls,
                                           &Controls::mouse_button));
  glutDisplayFunc (display_proxy);
  glutReshapeFunc (reshape_proxy);
  glutKeyboardFunc (key_proxy);
  glutKeyboardUpFunc (key_up_proxy);
  glutSpecialFunc (special_key_proxy);
  glutSpecialUpFunc (special_key_up_proxy);

  // Mouse Callbacks
  glutMouseFunc (mouse_button_proxy); // Buttons
  glutMotionFunc (mouse_motion_proxy); // Drag
  glutPassiveMotionFunc (mouse_motion_proxy); // Motion (no buttons pressed)
  glutIdleFunc (animate_proxy);
}

void Vamos_World::
Gl_World::set_attributes ()
{
  // Enable depth testing for hidden line removal
  glEnable (GL_DEPTH_TEST);
  // Allow transparency.
  glEnable (GL_BLEND);
  glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  // Allow textures.
  glEnable (GL_TEXTURE_2D);
  // Allow viewports.
  glEnable (GL_SCISSOR_TEST);
  // Allow masking.
  glEnable (GL_STENCIL_TEST);
  glClearStencil (0);
  // Allow lighting.
  glEnable (GL_LIGHTING);
  glEnable (GL_LIGHT0);
  // Create a Directional Light Source (x, y, z, w).
  GLfloat position [] = { 0.0, -1.0, 1.0, 0.0 };
  glLightfv (GL_LIGHT0, GL_POSITION, position);
  // Set the ambient light (R, G, B, A).
  GLfloat ambient [] = { 0.7, 0.7, 0.7, 1.0 };
  glLightfv (GL_LIGHT0, GL_AMBIENT, ambient);
  glClearColor (0.32, 0.65, 0.34, 0.0);
}

void Vamos_World::
Gl_World::add_car (Vamos_Body::Car* car)
{
  if (car == 0) return;
  World::add_car (car);
  m_view = BODY_VIEW;
  set_paused (false);
}

// Read the definition file.
void Vamos_World::
Gl_World::read (std::string data_dir, 
                std::string world_file,
                std::string controls_file)
{
  // Remember the file name for re-reading.
  if (data_dir != "")
    {
      m_data_dir = data_dir;
    }
  if (world_file != "")
    {
      m_world_file = world_file;
    }
  if (controls_file != "")
    {
      m_controls_file = controls_file;
    }

  World_Reader w_reader (m_data_dir + m_world_file, this);
  Controls_Reader c_reader (m_data_dir + m_controls_file, this);
}

bool Vamos_World::
Gl_World::fill_tank (double, double)
{
  focused_car ()->car->fuel_tank ()->fill ();
  return true;
}

void Vamos_World::
Gl_World::set_paused (bool is_paused)
{
  m_paused = is_paused;
  m_paused ? glutIdleFunc (0) : glutIdleFunc (animate_proxy);
}

bool Vamos_World::
Gl_World::pause (double, double)
{
  set_paused (!m_paused);
  return true;
}

bool Vamos_World::
Gl_World::quit (double, double)
{
  delete mp_window;
  mp_window = 0;
  std::exit (0);
  return true;
}

bool Vamos_World::
Gl_World::read_car (double, double)
{
  focused_car ()->car->read ();
  mp_sounds->set_engine_sound (focused_car ()->car->engine_sound ());
  return true;
}

bool Vamos_World::
Gl_World::read_track (double, double)
{
  mp_track->read ();
  display ();
  return true;
}

bool Vamos_World::
Gl_World::read_world (double, double)
{
  read ();
  return true;
}

bool Vamos_World::
Gl_World::reset_car (double, double)
{
  World::reset ();
  return true;
}

bool Vamos_World::
Gl_World::restart_car (double, double)
{
  World::restart ();
  return true;
}

bool Vamos_World::
Gl_World::start_engine (double, double)
{
  focused_car ()->car->start_engine ();
  return true;
}

bool Vamos_World::
Gl_World::cycle_view (double, double)
{
  switch (m_view)
    {
    case BODY_VIEW:
      m_view = MAP_VIEW;
      glClearColor (0.32, 0.65, 0.34, 0.0);
      break;
    case MAP_VIEW:
      m_view = WORLD_VIEW;
      break;
    case WORLD_VIEW:
    default:
      m_view = BODY_VIEW;
      break;
    }

  return true;
}

bool Vamos_World::
Gl_World::initial_shift_up (double, double)
{
  if (focused_car ()->car->gear () == 0)
    {
      focused_car ()->car->shift_up ();
      return true;
    }
  return false;
}

bool Vamos_World::
Gl_World::initial_shift_down (double, double)
{
  if (focused_car ()->car->gear () == 0)
    {
      focused_car ()->car->shift_down ();
      return true;
    }
  return false;
}

bool Vamos_World::
Gl_World::shift_up (double, double)
{
  if (focused_car ()->car->gear () != 0)
    {
      focused_car ()->car->shift_up ();
      return true;
    }
  return false;
}

bool Vamos_World::
Gl_World::shift_down (double, double)
{
  if (focused_car ()->car->gear () != 0)
    {
      focused_car ()->car->shift_down ();
      return true;
    }
  return false;
}

bool Vamos_World::
Gl_World::initial_shift_up_disengage (double, double time)
{
  if (focused_car ()->car->gear () == 0)
    {
      focused_car ()->car->disengage_clutch (time);
      focused_car ()->car->shift_up ();
      return true;
    }
  return false;
}

bool Vamos_World::
Gl_World::initial_shift_down_disengage (double, double time)
{
  if (focused_car ()->car->gear () == 0)
    {
      focused_car ()->car->disengage_clutch (time);
      focused_car ()->car->shift_down ();
      return true;
    }
  return false;
}

bool Vamos_World::
Gl_World::shift_up_disengage (double, double time)
{
  if (focused_car ()->car->gear () != 0)
    {
      if (!focused_car ()->car->fast_shift ())
        {
          focused_car ()->car->disengage_clutch (time);
        }
      focused_car ()->car->shift_up ();
      return true;
    }
  return false;
}

bool Vamos_World::
Gl_World::shift_down_disengage (double, double time)
{
  if (focused_car ()->car->gear () != 0)
    {
      if (!focused_car ()->car->fast_shift ())
        {
          focused_car ()->car->disengage_clutch (time);
        }
      focused_car ()->car->shift_down ();
      return true;
    }
  return false;
}

bool Vamos_World::
Gl_World::initial_clutch (double value, double)
{
  if (focused_car ()->car->last_gear () == 0)
    {
      focused_car ()->car->clutch (value);
      return true;
    }
  return false;
}

bool Vamos_World::
Gl_World::clutch (double value, double)
{
  if (focused_car ()->car->last_gear () != 0)
    {
      focused_car ()->car->clutch (value);
      return true;
    }
  return false;
}

bool Vamos_World::
Gl_World::initial_engage_clutch (double, double time)
{
  if (focused_car ()->car->last_gear () == 0)
    {
      focused_car ()->car->engage_clutch (time);
      return true;
    }
  return false;
}

bool Vamos_World::
Gl_World::initial_disengage_clutch (double, double time)
{
  if (focused_car ()->car->last_gear () == 0)
    {
      focused_car ()->car->disengage_clutch (time);
      return true;
    }
  return false;
}

bool Vamos_World::
Gl_World::engage_clutch (double, double time)
{
  if (focused_car ()->car->last_gear () != 0)
    {
      focused_car ()->car->engage_clutch (time);
      return true;
    }
  return false;
}

bool Vamos_World::
Gl_World::disengage_clutch (double, double time)
{
  if (focused_car ()->car->last_gear () != 0)
    {
      focused_car ()->car->disengage_clutch (time);
      return true;
    }
  return false;
}

bool Vamos_World::
Gl_World::gas (double value, double time)
{
  focused_car ()->car->gas (value, time);
  return false;
}

bool Vamos_World::
Gl_World::brake (double value, double time)
{
  focused_car ()->car->brake (value, time);
  return false;
}

bool Vamos_World::
Gl_World::steer (double value, double time)
{
  focused_car ()->car->steer (value, time);
  return false;
}

bool Vamos_World::
Gl_World::steer_left (double value, double time)
{
  focused_car ()->car->steer (value, time);
  return true;
}

bool Vamos_World::
Gl_World::steer_right (double value, double time)
{
  focused_car ()->car->steer (-value, time);
  return true;
}

bool Vamos_World::
Gl_World::pan_left (double value, double time)
{
  focused_car ()->car->pan (value, time);
  return true;
}

bool Vamos_World::
Gl_World::pan_right (double value, double time)
{
  focused_car ()->car->pan (-value, time);
  return true;
}

bool Vamos_World::
Gl_World::focus_previous_car (double, double)
{
  focus_other_car (-1);
  return true;
}

bool Vamos_World::
Gl_World::focus_next_car (double, double)
{
  focus_other_car (1);
  return true;
}

void Vamos_World::
Gl_World::animate ()
{
  if (focused_car () != 0)
    {
      mp_controls->update ();

      const int repeat = mp_timer->steps_per_frame ();
      const double dt = mp_timer->get_substep ();
      for (int loop = 0; loop < repeat; loop++)
        {
          for (std::vector <Car_Information>::iterator it = m_cars.begin ();
               it != m_cars.end ();
               it++)
            {
              it->car->propagate (dt);
              interact (it->car, it->segment_index);

              for (std::vector <Car_Information>::iterator 
                     other = m_cars.begin ();
                   other != m_cars.end ();
                   other++)
                {
                  if (other != it)
                    {
                      collide (it->car, other->car);
                    }
                }
            }
        }
    }

  if ((focused_car () != 0) && (mp_sounds != 0))
    {
      double tire_slide = 0.0;
      double grass_speed = 0.0;
      double gravel_speed = 0.0;
      double scrape_speed = 0.0;
      double hard_crash_speed = 0.0;
      double soft_crash_speed = 0.0;

      for (std::vector <Contact_Info>::const_iterator 
             it = m_contact_info.begin ();
           it != m_contact_info.end ();
           it++)
        {
          if (it->car != focused_car ()->car) continue;

          if ((it->track_material == Material::ASPHALT)
              || (it->track_material == Material::CONCRETE)
              || (it->track_material == Material::METAL))
          {
            if (it->car_material == Material::RUBBER)
              {
                tire_slide = focused_car ()->car->slide ();
              }
            else if (it->car_material == Material::METAL)
              {
                scrape_speed = it->parallel_speed;
                hard_crash_speed = it->perpendicular_speed;
              }
          }
          else if (it->track_material == Material::GRASS)
            {
              grass_speed = it->parallel_speed;
            }
          else if (it->track_material == Material::GRAVEL)
            {
              gravel_speed = it->parallel_speed;
            }
          else if (it->track_material == Material::RUBBER)
            {
              soft_crash_speed = it->perpendicular_speed;
            }
        }

      mp_sounds->play_engine_sound (focused_car ()->car->engine_pitch (),
                                    focused_car ()->car->engine_volume ());

      mp_sounds->play_tire_squeal_sound (tire_slide);
      mp_sounds->play_grass_sound (grass_speed);
      mp_sounds->play_gravel_sound (gravel_speed);
      mp_sounds->play_scrape_sound (scrape_speed);
      mp_sounds->play_wind_sound ((focused_car ()->car->chassis ().
                                   cm_velocity () 
                                   - mp_atmosphere->velocity ()).abs ());
      mp_sounds->play_hard_crash_sound (hard_crash_speed);
      mp_sounds->play_soft_crash_sound (soft_crash_speed);

      mp_sounds->update ();
    }

  for (std::vector <Car_Information>::iterator it = m_cars.begin ();
       it != m_cars.end ();
       it++)
    {
      const double distance = 
        mp_track->distance (it->car->chassis ().position (), it->segment_index);
      const int sector = mp_track->sector (distance);
      it->timing.update (mp_timer->get_time_step (), distance, sector);
    }

  m_contact_info.clear ();

  mp_timer->add_frame ();

  glutPostRedisplay ();
}

void Vamos_World::
Gl_World::show_full_window ()
{
  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();

  glViewport (0, 0, mp_window->width (), mp_window->height ());
  glScissor (0, 0, mp_window->width (), mp_window->height ());
  glStencilFunc (GL_ALWAYS, 1, 1);
  glStencilOp (GL_KEEP, GL_KEEP, GL_KEEP);

  glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
}

void Vamos_World::
Gl_World::draw_mirror_views ()
{
  if (focused_car ()->car->pan () != 0.0) return;

  const int mirrors = focused_car ()->car->get_n_mirrors ();
  for (int i = 0; i < mirrors; i++)
    {
      focused_car ()->car->draw_rear_view (mp_window->aspect (), i);
      
      glMatrixMode (GL_MODELVIEW);
      
      // Enable the rearview mirror mask.
      glStencilFunc (GL_EQUAL, 1, 1);

      // Front and back are reversed in the mirrors, so Cull front
      // faces while we're drawing mirror views.
      glPushAttrib (GL_POLYGON_BIT);
      {
        glCullFace (GL_FRONT);
        mp_track->draw_sky (focused_car ()->car->view_position ());
        mp_track->draw ();
        draw_cars (false, false);
      }
      glPopAttrib ();
    }
}

void Vamos_World::
Gl_World::set_car_view (Vamos_Body::Car* car)
{
  assert (car != 0);
  car->set_perspective (mp_window->aspect ());
  car->view (0.0);
}

void Vamos_World::
Gl_World::set_map_view ()
{
  double min_x = mp_track->min_x ();
  double max_x = mp_track->max_x ();
  double min_y = mp_track->min_y ();
  double max_y = mp_track->max_y ();
      
  // Adjust the mins and maxes to keep the correct aspect ratio of the
  // track regardless of the window's size.
  double track_aspect = (max_x - min_x) / (max_y - min_y);
  double ratio = track_aspect / mp_window->aspect ();

  if (ratio < 1.0)
    {
      // The window is wider than the track.  Stretch the x-dimension.
      min_x /= ratio;
      max_x /= ratio;
    }
  else
    {
      // The track is wider than the window.  Stretch the y-dimension.
      min_y *= ratio;
      max_y *= ratio;
    }
  glOrtho (min_x, max_x, min_y, max_y, -1000, 1000);
}

void Vamos_World::
Gl_World::set_world_view ()
{
  gluPerspective (10, mp_window->aspect (), 10.0, 1000.0);
  Three_Vector pos = focused_car ()->car->chassis ().cm_position ();
  gluLookAt (40.0, 40.0, 10.0, // camera position 
             pos [0], pos [1], pos [2],  // target 
             0.0, 0.0, 1.0); // up direction
}

void Vamos_World::
Gl_World::draw_track (bool draw_sky, const Three_Vector& view_position)
{
  glMatrixMode (GL_MODELVIEW);
  if (draw_sky)
    {
      assert (focused_car () != 0);
      mp_track->draw_sky (view_position);
    }
  mp_track->draw ();
}

void Vamos_World::
Gl_World::draw_cars (bool draw_interior, bool draw_focused_car)
{
  for (std::vector <Car_Information>::iterator it = m_cars.begin ();
       it != m_cars.end ();
       it++)
    {
      assert (it->car != 0);
      if (it->car != focused_car ()->car)
        {
          it->car->draw ();
          if (draw_interior)
            {
              it->car->draw_interior ();
            }
        }
    }
  if (draw_focused_car)
    {
      focused_car ()->car->draw ();
      if (draw_interior)
        {
          focused_car ()->car->draw_interior ();
        }
    }
}

void Vamos_World::
Gl_World::show_scene ()
{
  glFlush ();
  glutSwapBuffers ();
}

void Vamos_World::
Gl_World::display ()
{
  show_full_window ();

  switch (m_view)
    {
    case BODY_VIEW:
      set_car_view (focused_car ()->car);
      draw_track (true, focused_car ()->car->view_position ());
      draw_cars (true);
      draw_timing_info ();
      draw_mirror_views ();
      break;
    case MAP_VIEW:
      set_map_view ();
      draw_track (false);
      if (focused_car () != 0)
        {
          draw_cars (false);
        }
      break;
    case WORLD_VIEW:
      set_world_view ();
      draw_track (true, Three_Vector (40.0, 40.0, 10.0));
      draw_cars (true);
      break;
    }

  show_scene ();
}

void Vamos_World::
Gl_World::reshape (int width, int height)
{
  mp_window->resize (width, height);
  mp_controls->set_window_size (width, height);
  if (focused_car () != 0)
    {
      focused_car ()->car->make_rear_view_mask (width, height);
    }
}

void Vamos_World::
Gl_World::draw_string (const std::string& str, double x, double y)
{
  glRasterPos2d (x, y);
  for (std::string::const_iterator it = str.begin (); it != str.end (); it++)
    {
      glutBitmapCharacter (GLUT_BITMAP_8_BY_13, *it);
    }
}


void Vamos_World::
Gl_World::draw_timing_info ()
{
  if (focused_car ()->car == 0) return;

  glDisable (GL_DEPTH_TEST);
  glDisable (GL_LIGHTING);
  glDisable (GL_TEXTURE_2D);

  glColor3f (1.0, 1.0, 1.0);
  
  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();
  glMatrixMode (GL_MODELVIEW);
  glLoadIdentity ();
  
  std::ostringstream b_stream;
  double x = 5.5;

  gluOrtho2D (0, 10, 0, 10);

  const Timing_Info& timing = focused_car ()->timing;
  b_stream << "Lap Time " << format_time (timing.get_lap_time ());
  draw_string (b_stream.str (), x, 1.4);

  b_stream.str ("");
  b_stream << "    Last " << format_time (timing.get_previous_lap_time ()) 
           << " " 
           << format_time_difference (timing.get_lap_time_difference ());
  draw_string (b_stream.str (), x, 1.0);

  b_stream.str ("");
  b_stream << "    Best " << format_time (timing.get_best_lap_time ());
  draw_string (b_stream.str (), x, 0.6);

  b_stream.str ("");
  b_stream << 1.0 / mp_timer->get_time_step () << " frame/s";
  draw_string (b_stream.str (), x, 0.2);

  x = 7.5;

  int sector = timing.get_sector ();
  b_stream.str ("");
  b_stream << "   Sector ";
  if (sector != -1)
    {
      b_stream << sector + 1 << " " 
               << format_time (timing.get_sector_time ());
    }
  draw_string (b_stream.str (), x, 1.4);

  b_stream.str ("");
  b_stream << "       Best "; 
  if (sector != -1)
    {
      b_stream << format_time (timing.get_best_sector_time (sector));
    }
  draw_string (b_stream.str (), x, 1.0);

  b_stream.str ("");
  b_stream << "Last Sector ";
  if (timing.get_previous_sector () != -1)
    {
      b_stream << format_time (timing.get_previous_sector_time ()) << "  " 
               << format_time_difference 
        (timing.get_previous_sector_time_difference ());
    }
  draw_string (b_stream.str (), x, 0.6);

  b_stream.str ("");
  b_stream << "Distance " << int (timing.get_distance ()) << " m";
  draw_string (b_stream.str (), x, 0.2);

  glEnable (GL_DEPTH_TEST);
  glEnable (GL_LIGHTING);
  glEnable (GL_TEXTURE_2D);
}

void Vamos_World::
Gl_World::start ()
{
  glutMainLoop ();
}

void Vamos_World::
Gl_World::set_focused_car (size_t index)
{
  World::set_focused_car (index);

  if (focused_car () != 0)
    {
      mp_sounds->set_engine_sound (focused_car ()->car->engine_sound ());
      focused_car ()->car->make_rear_view_mask (mp_window->width (), 
                                                mp_window->height ());
    }
}

//* Class World_Reader

//** Constructor
Vamos_World::
World_Reader::World_Reader (std::string file_name, Gl_World* world) 
  : mp_world (world)
{
  read (file_name);
}

void Vamos_World::
World_Reader::on_start_tag (const Vamos_Geometry::XML_Tag& tag)
{
  //  std::cout << "start " << tag.get_label () << std::endl;
  m_tag = tag.get_label ();
  m_path = m_path + '/' + m_tag;
}

void Vamos_World::
World_Reader::on_end_tag (const Vamos_Geometry::XML_Tag& tag)
{
  //  std::cerr << "end " << tag.get_label () << std::endl;
  m_tag = tag.get_label ();
  m_path = m_path.substr (0, m_path.find_last_of ("/"));
}

void Vamos_World::
World_Reader::on_data (std::string data_string)
{
  std::string data = remove_leading_space (data_string);
  if (data.size () == 0)
    {
      return;
    }
  std::istringstream is (data.c_str ());

  if (m_path == "/world/gravity")
    {
      double grav;
      is >> grav;
      mp_world->gravity (grav);
    }
  else if (m_path == "/world/maximum-time-step")
    {
      double step;
      is >> step;
      mp_world->set_max_time_step (step);
    }
  else if (m_path == "/world/atmosphere/density")
    {
      double density;
      is >> density;
      mp_world->mp_atmosphere->density (density);
    }
  else if (m_path == "/world/atmosphere/velocity")
    {
      Three_Vector velocity;
      is >> velocity;
      mp_world->mp_atmosphere->velocity (velocity);
    }
  else if (m_path == "/world/lighting/source-position")
    {
      Three_Vector pos;
      is >> pos;
      GLfloat position [] = { pos [0], pos [1], pos [2], 0.0 };
      glLightfv (GL_LIGHT0, GL_POSITION, position);
    }
  else if (m_path == "/world/lighting/ambient")
    {
      Three_Vector amb;
      is >> amb;
      GLfloat ambient [] = { amb [0], amb [1], amb [2], 1.0 };
      glLightfv (GL_LIGHT0, GL_AMBIENT, ambient);
    }
}

//* Class Controls_Reader

//** Constructor
Vamos_World::
Controls_Reader::Controls_Reader (std::string file_name, Gl_World* world) 
  : mp_world (world),
    mp_controls (world->mp_controls)
{
  m_function_map ["pause"] = &Gl_World::pause;
  m_function_map ["quit"] = &Gl_World::quit;
  m_function_map ["read track"] = &Gl_World::read_track;
  m_function_map ["read world"] = &Gl_World::read_world;
  m_function_map ["cycle view"] = &Gl_World::cycle_view;

  m_function_map ["reset car"] = &Gl_World::reset_car;
  m_function_map ["restart car"] = &Gl_World::restart_car;
  m_function_map ["read car"] = &Gl_World::read_car;
  m_function_map ["start engine"] = &Gl_World::start_engine;
  m_function_map ["fill tank"] = &Gl_World::fill_tank;

  m_function_map ["initial shift up"] = &Gl_World::initial_shift_up;
  m_function_map ["initial shift down"] = &Gl_World::initial_shift_down;
  m_function_map ["shift up"] = &Gl_World::shift_up;
  m_function_map ["shift down"] = &Gl_World::shift_down;
  m_function_map ["initial shift up disengage"] 
    = &Gl_World::initial_shift_up_disengage;
  m_function_map ["initial shift down disengage"] 
    = &Gl_World::initial_shift_down_disengage;
  m_function_map ["shift up disengage"] 
    = &Gl_World::shift_up_disengage;
  m_function_map ["shift down disengage"] 
    = &Gl_World::shift_down_disengage;
  m_function_map ["initial engage clutch"] 
    = &Gl_World::initial_engage_clutch;
  m_function_map ["initial disengage clutch"] 
    = &Gl_World::initial_disengage_clutch;
  m_function_map ["engage clutch"] = &Gl_World::engage_clutch;
  m_function_map ["disengage clutch"] = &Gl_World::disengage_clutch;
  m_function_map ["initial clutch"] = &Gl_World::initial_clutch;
  m_function_map ["clutch"] = &Gl_World::clutch;

  m_function_map ["steer"] = &Gl_World::steer;
  m_function_map ["steer right"] = &Gl_World::steer_right;
  m_function_map ["steer left"] = &Gl_World::steer_left;
  m_function_map ["gas"] = &Gl_World::gas;
  m_function_map ["brake"] = &Gl_World::brake;

  m_function_map ["pan left"] = &Gl_World::pan_left;
  m_function_map ["pan right"] = &Gl_World::pan_right;

  m_function_map ["focus previous car"] = &Gl_World::focus_previous_car;
  m_function_map ["focus next car"] = &Gl_World::focus_next_car;

  read (file_name);
}

void Vamos_World::
Controls_Reader::on_start_tag (const Vamos_Geometry::XML_Tag& tag)
{
  //  std::cout << "start " << tag.get_label () << std::endl;
  m_tag = tag.get_label ();
  m_path = m_path + '/' + tag.get_label ();

  if (m_tag == "bind")
    {
      m_function = "";
      m_control = 0;
      m_direction = DOWN;
      m_factor = 1.0;
      m_offset = 0.0;
      m_minimum = 0.0;
      m_minimum_position = 0.0;
      m_maximum = 1.0;
      m_maximum_position = 1.0;
      m_time = 0.0;
      m_deadband = 0.0;
    }
}

void Vamos_World::
Controls_Reader::on_end_tag (const Vamos_Geometry::XML_Tag& tag)
{
  //  std::cout << "end " << tag.get_label () << std::endl;
  m_tag = tag.get_label ();
  m_path = m_path.substr (0, m_path.find_last_of ("/"));

  if (m_tag == "up")
    {
      m_direction = UP;
    }
  else if (m_tag == "down")
    {
      m_direction = DOWN;
    }
  else if (m_tag == "bind")
    {
      std::map <std::string, Controls::Callback_Func>
        ::iterator it = m_function_map.find (m_function);

      if (it == m_function_map.end ())
        {
          throw Unknown_Function (m_function);
        }
      // Register the callback.
      switch (m_type)
        {
        case KEY:
          if (m_direction == UP)
            {
              if (m_special)
                {
                  mp_controls->bind_special_key_up (m_control, 
                                                    mp_world, 
                                                    it->second,
                                                    m_time);
                }
              else
                {
                  mp_controls->bind_key_up (m_control, 
                                            mp_world, 
                                            it->second,
                                            m_time);
                }
            }
          else // DOWN is the default.
            {
              if (m_special)
                {
                  mp_controls->bind_special_key_down (m_control, 
                                                      mp_world, 
                                                      it->second,
                                                      m_time);
                }
              else
                {
                  mp_controls->bind_key_down (m_control, 
                                              mp_world, 
                                              it->second,
                                              m_time);
                }
            }
          break;
        case JOYSTICK_BUTTON:
          if (m_direction == UP)
            {
              mp_controls->bind_button_up (m_control, mp_world, 
                                           it->second, m_time);
            }
          else // DOWN is the default.
            {
              mp_controls->bind_button_down (m_control, mp_world, 
                                             it->second, m_time);
            }
          break;
        case JOYSTICK_AXIS:
          mp_controls->bind_axis (m_control, mp_world, it->second,
                                  m_factor, m_offset, m_minimum, m_deadband);
          break;
        case MOUSE_BUTTON:
          if (m_direction == UP)
            {
              mp_controls->bind_mouse_button_up (m_control, mp_world, 
                                                 it->second, m_time);
            }
          else
            {
              mp_controls->bind_mouse_button_down (m_control, mp_world, 
                                                   it->second, m_time);
            }
          break;
        case MOUSE_MOTION:
          mp_controls->bind_mouse_motion (m_control, mp_world, it->second,
                                          m_minimum, m_minimum_position,
                                          m_maximum, m_maximum_position);
          break;
        default:
          assert (false);
        }
    }
}

int Vamos_World::
mouse_button_to_control (std::string button)
{
  int control = GLUT_LEFT_BUTTON;
  if ((button == "left") || (button == "Left") || (button == "LEFT"))
    {
      control = GLUT_LEFT_BUTTON;
    }
  if ((button == "middle") || (button == "Middle") || (button == "MIDDLE"))
    {
      control = GLUT_MIDDLE_BUTTON;
    }
  if ((button == "right") || (button == "Right") || (button == "RIGHT"))
    {
      control = GLUT_RIGHT_BUTTON;
    }
  return control;
}

int Vamos_World::
mouse_direction_to_control (std::string direction)
{
  int control = Controls::X;
  if ((direction == "x") || (direction == "X")) 
    {
      control = Controls::X;
    }
  if ((direction == "y") || (direction == "Y")) 
    {
      control = Controls::Y;
    }
  return control;
}

void Vamos_World::
Controls_Reader::on_data (std::string data_string)
{
  std::string data = remove_leading_space (data_string);
  if (data.size () == 0)
    {
      return;
    }
  std::istringstream is (data.c_str ());

  if (m_tag == "key")
    {
      m_type = KEY;
      std::string key;
      is >> key;
      m_control = translate_key (key, &m_special);
    }
  else if (m_tag == "button")
    {
      m_type = JOYSTICK_BUTTON;
      is >> m_control;
    }
  else if (m_tag == "mouse-button")
    {
      m_type = MOUSE_BUTTON;
      std::string button;
      is >> button;
      m_control = mouse_button_to_control (button);
    }
  else if (m_tag == "axis")
    {
      m_type = JOYSTICK_AXIS;
      is >> m_control;
    }
  else if (m_tag == "mouse-direction")
    {
      m_type = MOUSE_MOTION;
      std::string direction;
      is >> direction;
      m_control = mouse_direction_to_control (direction);
    }
  else if (m_tag == "function")
    {
      m_function = data;
    }
  else if (m_tag == "factor")
    {
      is >> m_factor;
    }
  else if (m_tag == "offset")
    {
      is >> m_offset;
    }
  else if (m_tag == "minimum")
    {
      is >> m_minimum;
    }
  else if (m_tag == "minimum-position")
    {
      is >> m_minimum_position;
    }
  else if (m_tag == "maximum")
    {
      is >> m_maximum;
    }
  else if (m_tag == "maximum-position")
    {
      is >> m_maximum_position;
    }
  else if (m_tag == "time")
    {
      is >> m_time;
    }
  else if (m_tag == "deadband")
    {
      is >> m_deadband;
    }
}


int Vamos_World::
translate_key (std::string key_name, bool* special)
{
  *special = false;

  // If key_name is a single character, return its integer value...
  if (key_name.size () == 1)
    {
      return int (key_name [0]);
    }

  // ...otherwise, try to interpret it as the name of a special key.

  // Downcase the string to do case-insensinive comparisons.
  std::transform (key_name.begin (), key_name.end (), key_name.begin (),
                  tolower);

  if (key_name == "escape")
    {
      return 27;
    }
  if (key_name == "delete")
    {
      return 127;
    }

  *special = true;

  if (key_name == "up")
    {
      return GLUT_KEY_UP;
    }
  if (key_name == "down")
    {
      return GLUT_KEY_DOWN;
    }
  if (key_name == "left")
    {
      return GLUT_KEY_LEFT;
    }
  if (key_name == "right")
    {
      return GLUT_KEY_RIGHT;
    }
  if (key_name == "insert")
    {
      return GLUT_KEY_INSERT;
    }
  if (key_name == "home")
    {
      return GLUT_KEY_HOME;
    }
  if (key_name == "end")
    {
      return GLUT_KEY_END;
    }
  if (key_name == "page up")
    {
      return GLUT_KEY_PAGE_UP;
    }
  if (key_name == "page down")
    {
      return GLUT_KEY_PAGE_DOWN;
    }
  if (key_name == "f1")
    {
      return GLUT_KEY_F1;
    }
  if (key_name == "f2")
    {
      return GLUT_KEY_F2;
    }
  if (key_name == "f3")
    {
      return GLUT_KEY_F3;
    }
  if (key_name == "f4")
    {
      return GLUT_KEY_F4;
    }
  if (key_name == "f5")
    {
      return GLUT_KEY_F5;
    }
  if (key_name == "f6")
    {
      return GLUT_KEY_F6;
    }
  if (key_name == "f7")
    {
      return GLUT_KEY_F7;
    }
  if (key_name == "f8")
    {
      return GLUT_KEY_F8;
    }
  if (key_name == "f9")
    {
      return GLUT_KEY_F9;
    }
  if (key_name == "f10")
    {
      return GLUT_KEY_F10;
    }
  if (key_name == "f11")
    {
      return GLUT_KEY_F11;
    }
  if (key_name == "f12")
    {
      return GLUT_KEY_F12;
    }

  assert (false);
  return 0;
}
