//  Sounds.cc - sound management
//
//  Vamos Automotive Simulator
//  Copyright (C) 2003--2004 Sam Varner
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <vamos/world/Sounds.h>
#include <vamos/geometry/Conversions.h>

#include <fstream>
#include <sstream>
#include <iostream>

using Vamos_Geometry::Sample;
using Vamos_Geometry::clip;

Vamos_World::
Sounds::Sounds () :
  mp_engine_sound (0),
  mp_tire_squeal_sound (0),
  mp_grass_sound (0),
  mp_gravel_sound (0),
  mp_scrape_sound (0),
  mp_wind_sound (0),
  mp_soft_crash_sound (0),
  mp_hard_crash_sound (0),
  mp_scheduler (0)
{
}

Vamos_World::
Sounds::~Sounds ()
{
  delete mp_soft_crash_sound;
  delete mp_hard_crash_sound;
  delete mp_wind_sound;
  delete mp_scrape_sound;
  delete mp_gravel_sound;
  delete mp_grass_sound;
  delete mp_tire_squeal_sound;

  delete mp_scheduler;
}

void Vamos_World::
Sounds::add_sample (std::string file, Sound_Type type, 
                    double volume, double pitch)
{
  Sample* sample = 0;
  file = m_data_dir + file;
  switch (type)
    {
    case TIRE_SQUEAL:
      sample = mp_tire_squeal_sound = new Sample (file, volume, pitch, 
                                                  true, true);
      break;
    case GRASS:
      sample = mp_grass_sound = new Sample (file, volume, pitch, true);
      break;
    case GRAVEL:
      sample = mp_gravel_sound = new Sample (file, volume, pitch, true);
      break;
    case SCRAPE:
      sample = mp_scrape_sound = new Sample (file, volume, pitch, true);
      break;
    case WIND:
      sample = mp_wind_sound = new Sample (file, volume, pitch, true);
      break;
    case SOFT_CRASH:
      sample = mp_soft_crash_sound = new Sample (file, volume, pitch, true);
      break;
    case HARD_CRASH:
      sample = mp_hard_crash_sound = new Sample (file, volume, pitch, true);
      break;
    default:
      assert (false);
    }
}

void Vamos_World::
Sounds::start_samples (int sample_rate, double buffer_duration)
{
  // Construct the scheduler.
  mp_scheduler = new slScheduler (sample_rate);
  mp_scheduler->setSafetyMargin (buffer_duration);

  if (mp_engine_sound != 0)
    {
      mp_engine_sound->schedule (mp_scheduler, true);
    }
  if (mp_tire_squeal_sound != 0)
    {
      mp_tire_squeal_sound->schedule (mp_scheduler, true);
    }
  if (mp_grass_sound != 0)
    {
      mp_grass_sound->schedule (mp_scheduler, true);
    }
  if (mp_gravel_sound != 0)
    {
      mp_gravel_sound->schedule (mp_scheduler, true);
    }
  if (mp_scrape_sound != 0)
    {
      mp_scrape_sound->schedule (mp_scheduler, true);
    }
  if (mp_wind_sound != 0)
    {
      mp_wind_sound->schedule (mp_scheduler, true);
    }
  if (mp_soft_crash_sound != 0)
    {
      mp_soft_crash_sound->schedule (mp_scheduler);
    }
  if (mp_hard_crash_sound != 0)
    {
      mp_hard_crash_sound->schedule (mp_scheduler);
    }
}

void Vamos_World::
Sounds::read (std::string data_dir, std::string sounds_file)
{
  delete mp_soft_crash_sound;
  delete mp_hard_crash_sound;
  delete mp_wind_sound;
  delete mp_scrape_sound;
  delete mp_gravel_sound;
  delete mp_grass_sound;
  delete mp_tire_squeal_sound;
    
  mp_soft_crash_sound = 0;
  mp_hard_crash_sound = 0;
  mp_wind_sound = 0;
  mp_scrape_sound = 0;
  mp_gravel_sound = 0;
  mp_grass_sound = 0;
  mp_tire_squeal_sound = 0;

  delete mp_scheduler;

  if (data_dir != "")
    {
      m_data_dir = data_dir;
    }
  if (sounds_file != "")
    {
      m_sounds_file = sounds_file;
    }
  Sounds_Reader reader (m_data_dir + m_sounds_file, this);
}

void Vamos_World::
Sounds::set_engine_sound (Sample* sound)
{
  if (mp_engine_sound != 0)
    {
      mp_engine_sound->stop ();
    }
  mp_engine_sound = sound;
  sound->schedule (mp_scheduler, true, 1);
}

void Vamos_World::
Sounds::play_engine_sound (double pitch, double volume)
{
  if (mp_engine_sound != 0)
    {
      if (volume > 0.0)
        {
          mp_engine_sound->play ();
          mp_engine_sound->pitch (pitch);
          mp_engine_sound->volume (volume);
        }
      else
        {
          mp_engine_sound->pause ();
        }
    }
}

void Vamos_World::
Sounds::play_tire_squeal_sound (double slide)
{
  double volume = clip (slide - 0.3, 0.0, 1.0);
  if (volume > 0.0)
    {
      mp_grass_sound->pause ();
      mp_gravel_sound->pause ();
      mp_scrape_sound->pause ();

      mp_tire_squeal_sound->play ();
      
      double pitch = 1.5 - slide / 2.0;
      if (pitch < 1.0)
        pitch = 1.0;

      mp_tire_squeal_sound->pitch (pitch);
      mp_tire_squeal_sound->volume (volume);
    }
  else
    {
      mp_tire_squeal_sound->pause ();
    }
}

void Vamos_World::
Sounds::play_grass_sound (double speed)
{
  double volume = clip (0.05 * speed, 0.0, 1.0);
  if (speed > 0.0)
    {
      mp_gravel_sound->pause ();

      mp_grass_sound->play ();
      mp_grass_sound->volume (volume);
    }
  else
    {
      mp_grass_sound->pause ();
    }
}

void Vamos_World::
Sounds::play_gravel_sound (double speed)
{
  double volume = clip (0.08 * speed, 0.0, 1.0);
  if (speed > 0.0)
    {
      mp_grass_sound->pause ();

      mp_gravel_sound->play ();
      mp_gravel_sound->volume (volume);
    }
  else
    {
      mp_gravel_sound->pause ();
    }
}

void Vamos_World::
Sounds::play_scrape_sound (double speed)
{
  double volume = clip (0.1 * speed, 0.0, 1.0);
  if (speed > 0.0)
    {
      mp_tire_squeal_sound->pause ();
      mp_grass_sound->pause ();
      mp_gravel_sound->pause ();

      mp_scrape_sound->play ();
      mp_scrape_sound->volume (volume);
    }
  else
    {
      mp_scheduler->pauseSample (mp_scrape_sound);
    }
}

void Vamos_World::
Sounds::play_wind_sound (double speed)
{
  double volume = clip (0.005 * speed, 0.0, 1.0);
  if (volume > 0.0)
    {
      mp_wind_sound->play ();
      mp_wind_sound->volume (volume);
    }
  else
    {
      mp_wind_sound->pause ();
    }
}

void Vamos_World::
Sounds::play_hard_crash_sound (double speed)
{
  double volume = clip (0.1 * (speed - 1.0), 0.0, 1.0);
  if (volume > 0.0)
    {
      mp_hard_crash_sound->volume (volume);
      mp_hard_crash_sound->play ();
    }
}

void Vamos_World::
Sounds::play_soft_crash_sound (double speed)
{
  double volume = clip (0.1 * (speed - 1.0), 0.0, 1.0);
  if (volume > 0.0)
    {
      mp_soft_crash_sound->volume (volume);
      mp_soft_crash_sound->play ();
    }
}

void Vamos_World::
Sounds::update ()
{ 
  mp_scheduler->update (); 
}

//* Class Sounds_Reader

//** Constructor
Vamos_World::
Sounds_Reader::Sounds_Reader (std::string file_name, Sounds* sounds) 
  : mp_sounds (sounds),
    m_rate (8000), // The default sample rate is 8 kHz.
    m_buffer_duration (0.2)
{
  read (file_name);
}

void Vamos_World::
Sounds_Reader::on_start_tag (const Vamos_Geometry::XML_Tag& tag)
{
  //  std::cout << tag.get_label () << std::endl;
  m_tag = tag.get_label ();
}

void Vamos_World::
Sounds_Reader::on_end_tag (const Vamos_Geometry::XML_Tag& tag)
{
  //  std::cout << tag.get_label () << std::endl;
  Sounds::Sound_Type type = Sounds::NONE;
  m_tag = tag.get_label ();
  if (m_tag == "tire-squeal")
    {
      type = Sounds::TIRE_SQUEAL;
    }
  else if (m_tag == "grass-sound")
    {
      type = Sounds::GRASS;
    }
  else if (m_tag == "gravel-sound")
    {
      type = Sounds::GRAVEL;
    }
  else if (m_tag == "scrape-sound")
    {
      type = Sounds::SCRAPE;
    }
  else if (m_tag == "wind-sound")
    {
      type = Sounds::WIND;
    }
  else if (m_tag == "soft-crash-sound")
    {
      type = Sounds::SOFT_CRASH;
    }
  else if (m_tag == "hard-crash-sound")
    {
      type = Sounds::HARD_CRASH;
    }
  else if (m_tag == "sounds")
    {
      mp_sounds->start_samples (m_rate, m_buffer_duration);
    }

  if (type != Sounds::NONE)
    {
      mp_sounds->add_sample (m_file, type, m_volume, m_pitch);
    }
}

void Vamos_World::
Sounds_Reader::on_data (std::string data_string)
{
  std::string data = remove_leading_space (data_string);
  if (data.size () == 0)
    {
      return;
    }
  std::istringstream is (data.c_str ());

  if (m_tag == "file")
    {
      m_file = data;
    }
  else if (m_tag == "pitch")
    {
      is >> m_pitch;
    }
  else if (m_tag == "volume")
    {
      is >> m_volume;
    }
  else if (m_tag == "sample-rate")
    {
      is >> m_rate;
    }
  else if (m_tag == "buffer-duration")
    {
      is >> m_buffer_duration;
    }
}
